part of '../flutterx_live_data.dart';

/// Some useful extensions
extension LiveDataExt<X> on LiveData<X> {
  /// Map this LiveData to a new LiveData which depends on this one
  MediatorLiveData<Y> transform<Y>(FutureOr<Y> Function(X) transform, {bool reactive = true}) =>
      MediatorLiveData<Y>(reactive: reactive)..addTransformSource<X>(this, transform);

  /// Creates a [Future] from this LiveData which completes when it's value is accepted which means it's initialized
  /// and (if specified) matches [accept] predicate.
  /// You can also set a [timeLimit] after which if the value has not yet been accepted, returned future will complete
  /// with [onTimeout] result if specified, otherwise with a [TimeoutException] error
  /// If [cancel] future is provided and it completes before value is accepted or timeOut is expired, returned future
  /// will complete with it's result
  Future<X> whenValue({
    bool Function(X)? accept,
    Duration? timeLimit,
    FutureOr<X> Function()? onTimeout,
    Future<X>? cancel,
  }) {
    final completer = Completer<X>();
    void _observe(X value) => accept == null || accept(value) ? completer.complete(value) : null;
    observeForever(_observe);
    cancel?.then(completer.complete, onError: completer.completeError);
    return (timeLimit == null ? completer.future : completer.future.timeout(timeLimit, onTimeout: onTimeout))
        .whenComplete(() => removeObserver(_observe));
  }
}

/// Shortcuts for transforming multiple LiveData sources into the mediator target
extension MediatorExt<Y> on MediatorMixin<Y> {
  /// Check [addTransformSources]
  void addTransformSource2<X1, X2>(LiveData<X1> x1, LiveData<X2> x2, FutureOr<Y> Function(X1, X2) transform) =>
      addTransformSources([x1, x2], (values) => transform(values[0], values[1]));

  /// Check [addTransformSources]
  void addTransformSource3<X1, X2, X3>(
          LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, FutureOr<Y> Function(X1, X2, X3) transform) =>
      addTransformSources([x1, x2, x3], (values) => transform(values[0], values[1], values[2]));

  /// Check [addTransformSources]
  void addTransformSource4<X1, X2, X3, X4>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<X4> x4,
          FutureOr<Y> Function(X1, X2, X3, X4) transform) =>
      addTransformSources([x1, x2, x3, x4], (values) => transform(values[0], values[1], values[2], values[3]));

  /// Check [addLiveTransformSources]
  Disposable addLiveTransformSource2<X1, X2>(
          LiveData<X1> x1, LiveData<X2> x2, LiveData<Y> Function(X1, X2) transform) =>
      addLiveTransformSources([x1, x2], (values) => transform(values[0], values[1]));

  /// Check [addLiveTransformSources]
  Disposable addLiveTransformSource3<X1, X2, X3>(
          LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<Y> Function(X1, X2, X3) transform) =>
      addLiveTransformSources([x1, x2, x3], (values) => transform(values[0], values[1], values[2]));

  /// Check [addLiveTransformSources]
  Disposable addLiveTransformSource4<X1, X2, X3, X4>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3, LiveData<X4> x4,
          LiveData<Y> Function(X1, X2, X3, X4) transform) =>
      addLiveTransformSources([x1, x2, x3, x4], (values) => transform(values[0], values[1], values[2], values[3]));

  /// Check [addLiveTransformSources]
  Disposable addLiveTransformSource5<X1, X2, X3, X4, X5>(LiveData<X1> x1, LiveData<X2> x2, LiveData<X3> x3,
          LiveData<X4> x4, LiveData<X5> x5, LiveData<Y> Function(X1, X2, X3, X4, X5) transform) =>
      addLiveTransformSources(
          [x1, x2, x3, x4, x5], (values) => transform(values[0], values[1], values[2], values[3], values[4]));
}

/// Shortcuts for observing multiple LiveData simultaneously
extension StateObserverExt on StateObserver {
  /// Observe within [StateObserver] context
  void observe<X>(LiveData<X> x, Observer<X> onChanged, {Dispatcher? dispatcher}) =>
      _observe<X>(_ObserverKey(this, onChanged), x, onChanged, dispatcher);

  /// Check [observeAll]
  void observe2<X1, X2>(
    LiveData<X1> x1,
    LiveData<X2> x2,
    void Function(X1, X2) onChanged, {
    Dispatcher? dispatcher,
  }) {
    final sources = [x1, x2];
    _observeAll(
        _CombinedObserverKey(sources, onChanged), sources, (values) => onChanged(values[0], values[1]), dispatcher);
  }

  /// Check [observeAll]
  void observe3<X1, X2, X3>(
    LiveData<X1> x1,
    LiveData<X2> x2,
    LiveData<X3> x3,
    void Function(X1, X2, X3) onChanged, {
    Dispatcher? dispatcher,
  }) {
    final sources = [x1, x2, x3];
    _observeAll(_CombinedObserverKey(sources, onChanged), sources,
        (values) => onChanged(values[0], values[1], values[2]), dispatcher);
  }

  /// Check [observeAll]
  void observe4<X1, X2, X3, X4>(
    LiveData<X1> x1,
    LiveData<X2> x2,
    LiveData<X3> x3,
    LiveData<X4> x4,
    void Function(X1, X2, X3, X4) onChanged, {
    Dispatcher? dispatcher,
  }) {
    final sources = [x1, x2, x3, x4];
    _observeAll(_CombinedObserverKey(sources, onChanged), sources,
        (values) => onChanged(values[0], values[1], values[2], values[3]), dispatcher);
  }

  /// Check [observeAll]
  void observe5<X1, X2, X3, X4, X5>(
    LiveData<X1> x1,
    LiveData<X2> x2,
    LiveData<X3> x3,
    LiveData<X4> x4,
    LiveData<X5> x5,
    void Function(X1, X2, X3, X4, X5) onChanged, {
    Dispatcher? dispatcher,
  }) {
    final sources = [x1, x2, x3, x4, x5];
    _observeAll(_CombinedObserverKey(sources, onChanged), sources,
        (values) => onChanged(values[0], values[1], values[2], values[3], values[4]), dispatcher);
  }

  /// Observe multiple sources simultaneously within this context. Check [StateObserverExt]
  void observeAll(
    Iterable<LiveData> sources,
    void Function(List values) onChanged, {
    Dispatcher? dispatcher,
  }) =>
      _observeAll(_CombinedObserverKey(sources.toList(growable: false), onChanged), sources, onChanged, dispatcher);
}

/// Check [observeAllForever]
CombinedObserverWrapper observe2Forever<X1, X2>(
  LiveData<X1> x1,
  LiveData<X2> x2,
  void Function(X1, X2) onChanged, {
  Dispatcher? dispatcher,
}) =>
    observeAllForever([x1, x2], (values) => onChanged(values[0], values[1]), dispatcher: dispatcher);

/// Check [observeAllForever]
CombinedObserverWrapper observe3Forever<X1, X2, X3>(
  LiveData<X1> x1,
  LiveData<X2> x2,
  LiveData<X3> x3,
  void Function(X1, X2, X3) onChanged, {
  Dispatcher? dispatcher,
}) =>
    observeAllForever([x1, x2, x3], (values) => onChanged(values[0], values[1], values[2]), dispatcher: dispatcher);

/// Check [observeAllForever]
CombinedObserverWrapper observe4Forever<X1, X2, X3, X4>(
  LiveData<X1> x1,
  LiveData<X2> x2,
  LiveData<X3> x3,
  LiveData<X4> x4,
  void Function(X1, X2, X3, X4) onChanged, {
  Dispatcher? dispatcher,
}) =>
    observeAllForever([x1, x2, x3, x4], (values) => onChanged(values[0], values[1], values[2], values[3]),
        dispatcher: dispatcher);

/// Check [observeAllForever]
CombinedObserverWrapper observe5Forever<X1, X2, X3, X4, X5>(
  LiveData<X1> x1,
  LiveData<X2> x2,
  LiveData<X3> x3,
  LiveData<X4> x4,
  LiveData<X5> x5,
  void Function(X1, X2, X3, X4, X5) onChanged, {
  Dispatcher? dispatcher,
}) =>
    observeAllForever(
        [x1, x2, x3, x4, x5], (values) => onChanged(values[0], values[1], values[2], values[3], values[4]),
        dispatcher: dispatcher);

/// Observe multiple sources simultaneously. Check [StateObserverExt]
CombinedObserverWrapper observeAllForever(
  Iterable<LiveData> sources,
  void Function(List values) onChanged, {
  Dispatcher? dispatcher,
}) {
  final result = <ObserverWrapper>[];
  _observeAllInternal(sources, onChanged,
      <V>(source, onChanged) => result.add(source.observeForever(onChanged, dispatcher: dispatcher)));
  return CombinedObserverWrapper._(result);
}

void _observeAllInternal(
  Iterable<LiveData> sources,
  void Function(List values) onChanged,
  void Function<V>(LiveData<V> source, Observer<V> onChanged) observe,
) {
  sources = sources.toList(growable: false);
  var ready = false;
  var notify = false;
  void observeAll(dynamic _) {
    List values;
    try {
      values = sources.map((data) => data._data).toList(growable: false);
    } catch (_) {
      return; // at least one source is not initialized
    }
    if (ready) {
      onChanged(values);
    } else {
      notify = true;
    }
  }

  for (final source in sources) observe(source, observeAll);
  if (notify) onChanged(sources.map((data) => data._data).toList(growable: false));
  ready = true;
}
