part of '../flutterx_live_data.dart';

/// A [LiveData] instance which can never changes and cannot have active observers. The value is immediately dispatched
/// Use this class only as fallback when other LiveData are not provided
/// [invalidate] has no effect on this LiveData if you want a LiveData which can be invalidated consider
/// using [LiveData.value]
/// E.g. LiveDataBuilder<String>(data: myData ?? const ImmutableLiveData('no value'), builder: ...);
@immutable
class ImmutableLiveData<T> implements LiveData<T> {
  static final Set<Object> _dispatching = {};

  @override
  Map<Observer<T>, ObserverWrapper<T>> get _observers => const {};

  @override
  final T _data;

  @override
  set _data(T value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get _version => 0;

  @override
  set _version(int value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get _activeCount => _dispatching.length;

  @override
  set _activeCount(int value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get version => 0;

  @override
  bool get initialized => true;

  @override
  bool get isActive => _dispatching.isNotEmpty;

  @override
  T get value => _data;

  @override
  T get valueOrNull => _data;

  /// Creates with value
  const ImmutableLiveData(this._data);

  @override
  void onActive() {}

  @override
  void onInactive() {}

  @override
  bool shouldUpdateValue(T oldValue, T newValue) => false;

  @override
  void postData(FutureOr<T> value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  void setData(T value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  void invalidate() {}

  @override
  ObserverWrapper<T> observeForever(Observer<T> onChanged, {Dispatcher? dispatcher}) =>
      addObserverWrapper(ObserverWrapper<T>(this, onChanged, dispatcher));

  @override
  void removeObserver(Observer<T> observer) => _dispatching.remove(observer);

  @override
  ObserverWrapper<T> addObserverWrapper(ObserverWrapper<T> observer) {
    final dispatcher = observer.dispatcher;
    if (dispatcher == null) {
      observer.onChanged(_data);
    } else {
      _dispatching.add(observer.onChanged);
      dispatcher.post(() => _dispatching.remove(observer.onChanged) ? observer.onChanged(_data) : null);
    }
    return observer;
  }

  @override
  void removeObserverWrapper(ObserverWrapper<T> observer) => removeObserver(observer.onChanged);

  @override
  void _changeActiveState(ObserverWrapper<T> observer, {required bool active}) {}

  @override
  void _dispatchingValue(ObserverWrapper<T>? initiator) {}

  @override
  String toString() => '${objectRuntimeType(this, 'LiveData')}($value)';
}

/// Same as [ImmutableLiveData] but without any value
@immutable
class LateImmutableLiveData<T> implements LiveData<T> {
  static late dynamic __data;

  @override
  Map<Observer<T>, ObserverWrapper<T>> get _observers => const {};

  @override
  T get _data => __data;

  @override
  set _data(T value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get _version => -1;

  @override
  set _version(int value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get _activeCount => 0;

  @override
  set _activeCount(int value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  int get version => -1;

  @override
  bool get initialized => false;

  @override
  bool get isActive => false;

  @override
  T get value => _data;

  @override
  T? get valueOrNull => null;

  /// Creates without value
  const LateImmutableLiveData();

  @override
  void onActive() {}

  @override
  void onInactive() {}

  @override
  bool shouldUpdateValue(T oldValue, T newValue) => false;

  @override
  void postData(FutureOr<T> value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  void setData(T value) => throw UnsupportedError('cannot set unmodifiable data');

  @override
  void invalidate() {}

  @override
  ObserverWrapper<T> observeForever(Observer<T> onChanged, {Dispatcher? dispatcher}) =>
      ObserverWrapper<T>(this, onChanged, dispatcher);

  @override
  void removeObserver(Observer<T> observer) {}

  @override
  ObserverWrapper<T> addObserverWrapper(ObserverWrapper<T> observer) => observer;

  @override
  void removeObserverWrapper(ObserverWrapper<T> observer) {}

  @override
  void _changeActiveState(ObserverWrapper<T> observer, {required bool active}) {}

  @override
  void _dispatchingValue(ObserverWrapper<T>? initiator) {}

  @override
  String toString() => '${objectRuntimeType(this, 'LiveData')}($value)';
}
